package xmldb;

import java.io.IOException;

import xmldb.Attributes;

/**
 * <p>This class parses the tokens delivered by the
 * {@link Scanner} and checks them for correctness.</p>
 *
 * <p>XML Technologies WS 15/16 - Assignments<br/>
 * Workgroup DBIS, University of Konstanz</p>
 */
public class Parser extends Constants {
  /** Verbose mode (used for testing). */
  private static final boolean VERBOSE = true;
  
  /** Events: callback instance. */
  private ParserEvents events;
  /** Scanner reference. */
  private Scanner scanner;

  /** Events: current tag. */
  private String currTag;
  /** Events: current attribute name. */
  private String currAttribute;
  /** Events: current elements attributes. */
  private Attributes currAttributes;
  /** Events: current tag state (opening / closing). */
  private boolean currOpening;

  /**
   * Main method.
   * @param args command line arguments
   * @throws IOException I/O exception
   */
  public static void main(String[] args) throws IOException {
    // get filename or use default
    String filename = args.length != 0 ? args[0] : INPUT;

    System.out.println("Parsing " + filename);

    // parse document
    Parser parser = new Parser(filename);
    parser.initTimer();
    parser.parse();
    
    parser.printTimer();
    parser.printMem();
  }

  /**
   * Constructor; initializes the parser.
   * @param filename the input file
   * @throws IOException I/O exception
   */
  public Parser(String filename) throws IOException {
    // creates a scanner reference
    scanner = new Scanner(filename);
  }

  /**
   * Constructor, specifying a notifier class.
   * @param filename the input file
   * @param xpe class instance which receives notifications
   * when the parser comes across certain XML elements
   * @throws IOException I/O exception
   */
  public Parser(String filename, ParserEvents xpe) throws IOException {
    this(filename);
    events = xpe;
  }

  /**
   * Parses all nodes.
   * @throws IOException I/O exception
   */
  public void parse() throws IOException {
    // start scanning
    scanner.next();
    document();
  }
    
  /**
   * Document node.
   * @throws IOException I/O exception
   */
  public void document() throws IOException {
    if(VERBOSE) verbose("document");
    element();

    // document was not completely parsed...
    if(!scanner.finished()) {
      error("XMLParser: Parsing interrupted at line " + 
        scanner.getLine() + ", column " + scanner.getColumn());
    }
  }

  /**
   * Element node.
   * @throws IOException I/O exception
   */
  private void element() throws IOException {
    if(VERBOSE) verbose("element");
    element1();
    element2();
  }

  /**
   * Start of an element node.
   * @throws IOException I/O exception
   */
  private void element1() throws IOException {
    if(VERBOSE) verbose("element1");
    consume(Token.L_BR);
    consume(Token.NAME);
    attributeList1();
  }

  /**
   * Rest of an element node.
   * @throws IOException I/O exception
   */
  private void element2() throws IOException {
    if(VERBOSE) verbose("element2");

    // TODO: Exercise 1, add functions/check token type
    
    switch(scanner.getToken()) {
    case R_BR:
      consume(Token.R_BR);
      content();
      endTag();
      break;
    case CLOSE_R_BR:
      consume(Token.CLOSE_R_BR);
      break;
    default:
      error("Unexpected token in element2()");
    }
  }

  /**
   * Beginning of an attribute list.
   * @throws IOException I/O exception
   */
  private void attributeList1() throws IOException {
    if(VERBOSE) verbose("attributeList1");

    // TODO: Exercise 1, add functions/check token type
    
    switch(scanner.getToken()) {
    case R_BR:
    case CLOSE_R_BR:
      break;
    case SPACE:
      consume(Token.SPACE);
      attributeList2();
      break;
    default:
      error("Unexpected token in attributeList1()");
    }
  }

  /**
   * End of an attribute list.
   * @throws IOException I/O exception
   */
  private void attributeList2() throws IOException {
    if(VERBOSE) verbose("attributeList2");

    // TODO: Exercise 1, add functions/check token type
    
    switch(scanner.getToken()) {
    case R_BR:
    case CLOSE_R_BR:
      break;
    case NAME:
      attribute();
      attributeList1();
      break;
    default:
      error("Unexpected token in attributeList2()");
  }
  }

  /**
   * Attribute.
   * @throws IOException I/O exception
   */
  private void attribute() throws IOException {
    if(VERBOSE) verbose("attribute");

    // TODO: Exercise 1, add functions/check token type
    
    consume(Token.NAME);
    optionalSpace();
    consume(Token.EQ);
    optionalSpace();
    consume(Token.QUOTE);
    consume(Token.ATT_VAL);
    consume(Token.QUOTE);
  }

  /**
   * End tag.
   * @throws IOException I/O exception
   */
  private void endTag() throws IOException {
    if(VERBOSE) verbose("endTag");

    // TODO: Exercise 1, add functions/check token type
    
    consume(Token.L_BR_CLOSE);
    consume(Token.NAME);
    optionalSpace();
    consume(Token.R_BR);
  }

  /**
   * Content node.
   * @throws IOException I/O exception
   */
  private void content() throws IOException {
    if(VERBOSE) verbose("content");

    // TODO: Exercise 1, add functions/check token type
    
    switch(scanner.getToken()) {
    case L_BR_CLOSE:
      break;
    case TEXT:
      consume(Token.TEXT);
      content();
      break;
    case L_BR:
      element();
      content();
      break;
    default:
      error("Unexpected token in content()");
  }
  }

  /**
   * Optional space.
   * @throws IOException I/O exception
   */
  private void optionalSpace() throws IOException {
    if(VERBOSE) verbose("optionalSpace");
    switch(scanner.getToken()) {
      case SPACE:
        consume(Token.SPACE);
        break;
      case R_BR:
      case EQ:
      case QUOTE:
        break;
      default:
        error("Unexpected token in optionalSpace()");
    }
  }

  /**
   * Consumes a token.
   * @param token the specified token
   * @throws IOException I/O exception
   */
  private void consume(Token token) throws IOException {
    if(token != scanner.getToken()) {
      // error found by parser - stopping
      error("Wrong token at line " + scanner.getLine() + ", column " +
        scanner.getColumn() + ": " + token +  " expected, " + 
        scanner.getToken() + " found (" + scanner.getContent() + ").");
      return;
    }

    if(VERBOSE) {
      StringBuilder sb = new StringBuilder("consume     ");
      sb.append(token.toString());
      while(sb.length() < 24) sb.append(' ');
      verbose(sb + scanner.getContent());
    }

    if(events != null) {
    	/*
         * TODO: Exercise 2, interpret the tokens and call the methods of the
         * {@link ParserEvents} interface:
         * 
         * - cache intermediary input when you encounter the following tokens:
         *   L_BR, L_BR_CLOSE, NAME, ATT_VAL
         * - send events when the following tokens are scanned:
         *   R_BR, CLOSE_R_BR or TEXT
         * 
         * You can use the variables currTag, currAttribute, currAttributes and
         * currOpening to temporarily save parsing results, or find your own
         * solution. - If no attributes are given, leave currAttributes = null
         * to save memory.
         */
    	
    	if(token == Token.L_BR) {
            // opening tag
            currOpening = true;
            currAttributes = null;
            currTag = null;
          } else if(token == Token.L_BR_CLOSE) {
            // closing tag
            currOpening = false;
            currTag = null;
          } else if(token == Token.NAME) {
            // name of element or attribute
            if(currTag == null) {
              currTag = scanner.getContent();
            } else {
              currAttribute = scanner.getContent();
            }
          } else if(token == Token.ATT_VAL) {
            // attribute value
            if(currAttributes == null) currAttributes = new Attributes();
            currAttributes.add(currAttribute, scanner.getContent());
          } else if(token == Token.R_BR) {
            // end of tag - send corresponding event
            if(currOpening) events.startElement(currTag, currAttributes);
            else events.endElement(currTag);
          } else if(token == Token.CLOSE_R_BR) {
            // end of empty tag - send startElement and endElement event
            events.startElement(currTag, currAttributes);
            events.endElement(currTag);
          } else if(token == Token.TEXT) {
            // send content event
            events.characters(scanner.getContent());
          }
    	/*
    	 * ************ solution END. 
    	 */
    }
    
    // scan next token
    scanner.next();
  }
}
